# -*- coding: utf-8 -*-

import os
import sys
import textwrap
from bs4 import BeautifulSoup
from colorama import Fore

from wallabag.api.api import ApiException
from wallabag.api.get_entry import GetEntry
from wallabag.commands.command import Command
from wallabag.entry import Entry


class ShowCommandParams():

    def __init__(self, entry_id, colors=True, html=False,
                 raw=False, image_links=False):
        self.entry_id = entry_id
        self.colors = colors
        self.html = html
        self.raw = raw
        self.image_links = image_links


class ShowCommand(Command):

    def __init__(self, config, params):
        self.config = config
        self.params = params

    def run(self):
        try:
            api = GetEntry(self.config, self.params.entry_id)
            entry = Entry(api.request().response)
        except ApiException as ex:
            return False, str(ex)

        article = entry.content
        if not self.params.html:
            article = self.__html2text(article)

        output = f"{entry.title}\n{self.__header_delimiter()}\n{article}"
        if not self.params.raw:
            output = self.__format_output(output)
        return True, output

    def __format_output(self, output):
        try:
            maxcol = os.get_terminal_size().columns
        except OSError:
            maxcol = sys.maxsize

        result = []
        for line in output.splitlines():
            result.extend(textwrap.wrap(line, maxcol) if line else [''])
        return "\n".join(result)

    def __html2text(self, html):
        soup = BeautifulSoup(html, "html.parser")
        self.__color_headers(soup)
        self.__color_bold(soup)
        self.__make_hr(soup)
        self.__replace_images(soup)
        return soup.text

    def __header_delimiter(self):
        try:
            return "".ljust(os.get_terminal_size().columns, '=')
        except OSError:
            return "\n"

    def __color_headers(self, soup):
        if self.params.colors:
            h1colors = Fore.BLUE
            h1colore = Fore.RESET
        else:
            h1colors = h1colore = ""
        for header in ['h1', 'h2', 'h3']:
            for h in soup.findAll(header):
                h.string = f"{h1colors}{h.string}{h1colore}"
                h.insert_before(self.__get_new_line_tag(soup, 2))
                h.insert_after(self.__get_new_line_tag(soup))
                h.unwrap()

    def __get_new_line_tag(self, soup, times=1):
        p = soup.new_tag('p')
        p.string = "\n" * times
        return p

    def __color_bold(self, soup):
        if self.params.colors:
            bcolors = Fore.RED
            bcolore = Fore.RESET
            for bold in ['b', 'strong']:
                for b in soup.findAll('b'):
                    b.string = f"{bcolors}{b.string}{bcolore}"

    def __make_hr(self, soup):
        try:
            hrstring = "".ljust(os.get_terminal_size().columns, '-')
        except OSError:
            hrstring = "-----"
        for hr in soup.findAll('hr'):
            replace = soup.new_tag('p')
            replace.string = hrstring
            hr.insert_after(replace)
            hr.unwrap()

    def __replace_images(self, soup):
        for img in soup.findAll('img'):
            replace = soup.new_tag('p')
            try:
                alt = f" \"{img['alt']}\""
            except KeyError:
                alt = ""
            links = ""
            if self.params.image_links:
                links = f" ({img['src']})"
            replace.string = f"[IMAGE{alt}{links}]\n"
            img.insert_after(replace)
            img.unwrap()



#!/usr/bin/env python3

import functools
import platform
import subprocess
from sys import exit

import click

from wallabag.commands.add import AddCommand, AddCommandParams
from wallabag.commands.delete import DeleteCommand, DeleteCommandParams
from wallabag.commands.list import ListCommand, ListParams, CountCommand
from wallabag.commands.show import ShowCommand, ShowCommandParams
from wallabag.commands.update import UpdateCommand, UpdateCommandParams
from wallabag.config import Configs
from wallabag.configurator import (
        ClientOption,
        Configurator,
        PasswordOption,
        SecretOption,
        Validator,
    )


@click.group()
@click.option('--config', help='Use custom configuration file')
@click.pass_context
def cli(ctx, config):
    # Workaround for default non-unicode encodings in the
    # Windows cmd and Powershell
    # -> Analyze encoding and set to utf-8
    if platform.system() == "Windows":
        codepage = subprocess.check_output(['chcp'], shell=True).decode()
        if "65001" not in codepage:
            subprocess.check_output(['chcp', '65001'], shell=True)

    ctx.obj = Configs(config)


def need_config(func):
    @functools.wraps(func)
    @click.pass_context
    def wrapper(ctx, *args, **kwargs):
        if not ctx.obj.is_valid():
            i = input(
                """Could not find a valid config.
Would you like to create it now? [Y/n]
""")
            if str.lower(i) in ["y", "yes", ""]:
                Configurator(ctx.obj).start()
            else:
                exit(0)
        func(*args, **kwargs)

    return wrapper


@cli.command()
@click.option('-s/-u', '--starred/--unstarred', default=None,
              help="Show only starred/unstarred entries.")
@click.option('-r/-n', '--read/--unread', default=None,
              help="Show only read/unread entries.")
@click.option('-a', '--all', default=False, is_flag=True,
              help="Show read as well as unread entries.")
@click.option('-o', '--oldest', default=False, is_flag=True,
              help="Show oldest matches instead of the newest.")
@click.option('-t', '--trim-output', default=False, is_flag=True,
              help="Trim the titles to fit the length of the cli.")
@click.option('-c', '--count', default=False, is_flag=True,
              help="Show a sum of matching entries.")
@click.option('-q', '--quantity', type=click.INT,
              help="Set the number of entries to show.")
@need_config
@click.pass_context
def list(ctx, starred, read, all, oldest, trim_output, count, quantity):
    """
    List the entries on the wallabag account.

    Gives a summary of entries in wallabag. Use options to filter the results.
    """
    config = ctx.obj
    if all:
        read = None
        starred = None

    params = ListParams(quantity, read, starred, oldest, trim_output)
    run_command(
            CountCommand(config, params) if count else
            ListCommand(config, params))


@cli.command()
@click.option('-c/-n', '--color/--no-color', default=True)
@click.option('-i', '--image-links', default=False, is_flag=True,
              help="Show image links in optimized output")
@click.option('-r', '--raw', default=False, is_flag=True,
              help="Disable wordwise trimming")
@click.option('-t', '--html', default=False, is_flag=True,
              help="Show the entry as html instead of optimized output for \
the cli.")
@click.argument('entry_id', required=True)
@need_config
@click.pass_context
def show(ctx, entry_id, color, html, raw, image_links):
    """
    Show the text of an entry.

    The ENTRY_ID can be found with `list` command.
    """
    run_command(ShowCommand(
        ctx.obj, ShowCommandParams(entry_id, color, html, raw, image_links)))


@cli.command()
@click.option('-q', '--quiet', default=False, is_flag=True,
              help="Hide the output if no error occurs.")
@click.argument('entry_id', required=True)
@need_config
@click.pass_context
def read(ctx, entry_id, quiet):
    """
    Toggle the read-status of an existing entry.

    This is an alias for `update --toggle-read <ENTRY_ID>` command.

    The ENTRY_ID can be found with `list` command.
    """
    params = UpdateCommandParams(entry_id)
    params.toggle_read = True
    params.quiet = quiet
    run_command(UpdateCommand(ctx.obj, params))


@cli.command()
@click.option('-q', '--quiet', default=False, is_flag=True,
              help="Hide the output if no error occurs.")
@click.argument('entry_id', required=True)
@need_config
@click.pass_context
def star(ctx, entry_id, quiet):
    """
    Toggle the starred-status of an existing entry.

    This is an alias for `update --toggle-starred <ENTRY_ID>` command.

    The ENTRY_ID can be found with `list` command.
    """
    params = UpdateCommandParams(entry_id)
    params.toggle_star = True
    params.quiet = quiet
    run_command(UpdateCommand(ctx.obj, params))


@cli.command()
@click.option('-t', '--title', default="", help="Add a custom title.")
@click.option('-r', '--read', default=False, is_flag=True,
              help="Mark as read.")
@click.option('-s', '--starred', default=False, is_flag=True,
              help="Mark as starred.")
@click.option('-q', '--quiet', default=False, is_flag=True,
              help="Hide the output if no error occurs.")
@click.argument('url', required=True)
@need_config
@click.pass_context
def add(ctx, url, title, read, starred, quiet):
    """Add a new entry to wallabag."""
    params = AddCommandParams(url, title, read, starred)
    run_command(AddCommand(ctx.obj, params), quiet)


@cli.command()
@click.option('-f', '--force', default=False, is_flag=True,
              help="Do not ask before deletion.")
@click.option('-q', '--quiet', default=False, is_flag=True,
              help="Hide the output if no error occurs.")
@click.argument('entry_id', required=True)
@need_config
@click.pass_context
def delete(ctx, entry_id, force, quiet):
    """
    Delete an entry from wallabag.

    The ENTRY_ID can be found with `list` command.
    """
    params = DeleteCommandParams(entry_id, force, quiet)
    run_command(DeleteCommand(ctx.obj, params))


@cli.command()
@click.option('-t', '--title', default="", help="Change the title.")
@click.option('-r', '--toggle-read', is_flag=True,
              help="Toggle the read status")
@click.option('-s', '--toggle-starred', is_flag=True,
              help="Toggle the starred status")
@click.option('-q', '--quiet', is_flag=True,
              help="Hide the output if no error occurs.")
@click.argument('entry_id', required=True)
@need_config
@click.pass_context
def update(ctx, entry_id, title, toggle_read, toggle_starred, quiet):
    """
        Toggle the read or starred status or change the title
        of an existing entry.

        The ENTRY_ID can be found with `list` command.
    """
    params = UpdateCommandParams(entry_id)
    params.new_title = title
    params.toggle_read = toggle_read
    params.toggle_star = toggle_starred
    params.quiet = quiet
    run_command(UpdateCommand(ctx.obj, params))


@cli.command()
@click.option('-c', '--check', is_flag=True,
              help="Check the config for errors.")
@click.option('-p', '--password', is_flag=True,
              help="Change the wallabag password.")
@click.option('-o', '--oauth', is_flag=True,
              help="Change the wallabag client credentials.")
@click.pass_context
def config(ctx, check, password, oauth):
    config = ctx.obj
    if check:
        (result, msg) = Validator(config).check()
        click.echo(msg)
        exit(result)
    options = []
    if password or oauth:
        if not config.is_valid():
            click.echo(
                "Invalid existing config.\
                        Therefore you have to enter all values.")
        else:
            if password:
                options.append(PasswordOption())
            if oauth:
                options.append(ClientOption())
                options.append(SecretOption())
    configurator = Configurator(config)
    while True:
        configurator.start(options)
        (result, msg, options) = Validator(config).check_oauth()
        if result or not options:
            click.echo(msg)
            exit(0)


def run_command(command, quiet=False):
    result, output = command.run()
    if not quiet and output:
        click.echo(output)
    if not result:
        exit(1)



# -*- coding: utf-8 -*-

from wallabag.api.api import ApiException
from wallabag.commands.command import Command
from wallabag.api.get_entry import GetEntry
from wallabag.api.update_entry import UpdateEntry, Params
from wallabag.entry import Entry


class UpdateCommandParams():
    entry_id = None
    toggle_read = False
    toggle_star = False
    new_title = None
    quiet = False

    def __init__(self, entry_id):
        self.entry_id = entry_id

    def validate(self):
        if not self.new_title and not self.toggle_read\
                and not self.toggle_star:
            return False
        return True


class UpdateCommand(Command):

    def __init__(self, config, params):
        self.config = config
        self.params = params or UpdateCommandParams()

    def run(self):
        params = self.params
        if not params.validate():
            return False, "Error: No parameter given."

        read_value = None
        star_value = None

        try:
            request = GetEntry(self.config, params.entry_id).request()
            entry = Entry(request.response)
            if params.toggle_read:
                read_value = not entry.read
            if params.toggle_star:
                star_value = not entry.starred
        except ApiException as ex:
            return False, str(ex)

        try:
            request = UpdateEntry(self.config, params.entry_id, {
                Params.TITLE: params.new_title,
                Params.STAR: star_value,
                Params.READ: read_value
            }).request()
            if not params.quiet:
                return True, "Entry successfully updated."
        except ApiException as ex:
            return False, str(ex)



# -*- coding: utf-8 -*-

import pytest

from wallabag.api.api import Response
from wallabag.api.get_list_entries import GetListEntries
from wallabag.commands.list import ListCommand, ListParams
from wallabag.config import Configs


class TestListCommand():

    def setup_method(self, method):
        self.config = Configs("/tmp/config")
        self.config.config.read_string("""
                [api]
                serverurl = url
                username = user
                password = pass
                [oauth2]
                client = 100
                secret = 100
                """)

    def test_empty_list(self, monkeypatch):

        def list_entries(self):
            text = '{ "_embedded": { "items": [] } }'
            return Response(200, text)

        monkeypatch.setattr(GetListEntries, 'request', list_entries)

        command = ListCommand(self.config)
        result, entries = command.run()
        assert result
        assert not entries

    @pytest.mark.parametrize('values', [
        ((0, 0), '1 title'), ((0, 1), '1 * title'),
        ((1, 0), '1 ✔ title'), ((1, 1), '1 ✔* title'),
        ])
    def test_entries_list(self, monkeypatch, values):

        def list_entries(self):
            text = '''
            { "_embedded": { "items": [
                {
                "id": 1,
                "title": "title",
                "content": "content",
                "url": "url",
                "is_archived": %s,
                "is_starred": %s
                }
            ] } }
            ''' % (values[0][0], values[0][1])
            return Response(200, text)

        monkeypatch.setattr(GetListEntries, 'request', list_entries)

        command = ListCommand(self.config)
        result, entries = command.run()
        assert result
        assert entries
        assert entries == values[1]

    def test_list(self, monkeypatch):

        def list_entries(self):
            text = '''
            { "_embedded": { "items": [
                { "id": 1, "title": "title", "content": "content",
                "url": "url", "is_archived": 0, "is_starred": 1},
                { "id": 2, "title": "title", "content": "content",
                "url": "url", "is_archived": 0, "is_starred": 1}
            ] } }
            '''
            return Response(200, text)

        monkeypatch.setattr(GetListEntries, 'request', list_entries)

        command = ListCommand(self.config)
        result, entries = command.run()
        assert result
        assert len(entries.split('\n')) == 2



# -*- coding: utf-8 -*-

from enum import Enum

from wallabag.api.api import Api, ApiMethod


class Params(Enum):
    COUNT = "count"
    FILTER_READ = "filter_read"
    FILTER_STARRED = "filter_starred"
    OLDEST = "oldest"


class GetListEntries(Api):

    API_METHOD = ApiMethod.LIST_ENTRIES

    class ApiParams(Enum):
        PER_PAGE = "perPage"
        OLDEST = "oldest"
        ARCHIVE = "archive"
        STARRED = "starred"

    class ApiValues(Enum):
        class OLDEST(Enum):
            ASC = "asc"

    def __init__(self, config, params):
        Api.__init__(self, config)
        self.params = params

    def _make_request(self, request):
        return self._request_get(request)

    def _get_api_url(self):
        return self._build_url(GetListEntries.API_METHOD)

    def _get_params(self):
        ApiParams = self.ApiParams
        ApiValues = self.ApiValues
        count = self.__get_count(self.params[Params.COUNT])
        api_params = {
            ApiParams.PER_PAGE.value: count
        }

        if Params.OLDEST in self.params and self.params[Params.OLDEST]:
            api_params[ApiParams.OLDEST.value] = ApiValues.OLDEST.value.ASC

        if Params.FILTER_READ in self.params:
            self._put_bool_param(
                    api_params, Params.FILTER_READ, ApiParams.ARCHIVE)

        if Params.FILTER_STARRED in self.params:
            self._put_bool_param(
                    api_params, Params.FILTER_STARRED, ApiParams.STARRED)

        return api_params

    def __get_count(self, count):
        return 0 if not count or count < 0 else count



# -*- coding: utf-8 -*-

from wallabag.api.api import Api, ApiMethod


class ApiVersion(Api):

    def __init__(self, config, custom_url=None):
        Api.__init__(self, config)
        self.custom_url = custom_url

    def _get_api_url(self):
        return self._build_url(ApiMethod.VERSION, self.custom_url)

    def _make_request(self, request):
        return self._request_get(request)



# -*- coding: utf-8 -*-

import os
import platform
import sys

from wallabag.api.api import ApiException
from wallabag.api.get_list_entries import GetListEntries, Params
from wallabag.commands.command import Command
from wallabag.entry import Entry


class ListParams():
    quantity = None
    filter_read = None
    filter_starred = None
    oldest = False
    trim = True

    def __init__(self, quantity=None, filter_read=None,
                 filter_starred=None, oldest=None, trim=None):
        self.quantity = quantity
        self.filter_read = filter_read
        self.filter_starred = filter_starred
        self.oldest = oldest
        self.trim = trim


class ListCommand(Command):

    def __init__(self, config, params=None):
        self.config = config
        self.params = params or ListParams()

    def run(self):
        try:
            api = GetListEntries(self.config, {
                Params.COUNT: self.__get_quantity(),
                Params.FILTER_READ: self.params.filter_read,
                Params.FILTER_STARRED: self.params.filter_starred,
                Params.OLDEST: self.params.oldest
            })
            entries = Entry.entrylist(
                    api.request().response['_embedded']["items"])
            return True, self.__print_entries(entries)
        except ApiException as ex:
            return False, str(ex)

    def __print_entries(self, entries):
        show_read_column, show_starred_column = self.__read_star_width(entries)
        entry_id_width = self.__entry_id_width(entries)
        maxwidth = self.__get_maxwidth()
        output = []
        if not self.params.oldest:
            entries = reversed(entries)
        for item in entries:
            entry_id = str(item.entry_id).rjust(entry_id_width)

            title = item.title
            read, starred = self.__read_star_char(item)
            line = entry_id
            if show_read_column or show_starred_column:
                line = line + " "
                if show_read_column:
                    line = line + read
                if show_starred_column:
                    line = line + starred

            line = f"{line} {title}"
            output.append(line[0:maxwidth])
        return '\n'.join(output)

    def __get_quantity(self):
        if self.params.quantity is None:
            try:
                return os.get_terminal_size().lines - 2
            except OSError:
                return sys.maxsize
        else:
            return self.params.quantity

    def __get_maxwidth(self):
        if self.params.trim:
            try:
                return os.get_terminal_size().columns
            except OSError:
                pass
        return sys.maxsize

    def __entry_id_width(self, entries):
        if len(entries) > 0:
            size_entry_id = len(str(entries[0].entry_id))
            entry_id_last = len(str(entries[len(entries) - 1].entry_id))
            if entry_id_last > size_entry_id:
                size_entry_id = entry_id_last
            return size_entry_id
        return 0

    def __read_star_width(self, entries):
        show_read_column = False
        show_starred_column = False
        for item in entries:
            if item.read:
                show_read_column = True
            if item.starred:
                show_starred_column = True
            if show_read_column and show_starred_column:
                break
        return (show_read_column, show_starred_column)

    def __read_star_char(self, item):
        read = " "
        if item.read:
            if platform.system() == "Windows":
                read = "r"
            else:
                read = "✔"

        starred = " "
        if item.starred:
            starred = "*"

        return (read, starred)


class CountCommand(Command):

    def __init__(self, config, params=None):
        self.config = config
        self.params = params or ListParams()

    def run(self):
        try:
            api = GetListEntries(self.config, {
                Params.COUNT: sys.maxsize,
                Params.FILTER_READ: self.params.filter_read,
                Params.FILTER_STARRED: self.params.filter_starred
            })
            response = api.request().response
        except ApiException as ex:
            return False, str(ex)

        return True, len(response["_embedded"]["items"])



"""
Wallabag entry.
"""


class Entry:
    """
    The entry class.
    """
    entry_id = 0
    title = ""
    content = ""
    url = ""
    read = False
    starred = False

    def __init__(self, item):
        self.entry_id = item['id']

        title = item['title']
        title = title.replace("\n", "")
        title = " ".join(title.split())
        self.title = title

        self.content = item['content']
        self.url = item['url']
        self.read = item['is_archived'] == 1
        self.starred = item['is_starred'] == 1

    def entrylist(items):
        """
        Creates Entry instances of multiple items and returns an Entry-list.
        """
        ret = list()
        for item in items:
            ret.append(Entry(item))
        return ret
